from __future__ import unicode_literals
from collections import Counter, namedtuple
from functools import wraps


def _merge_one(lists, tails):
    for list in lists:
        if not list: continue
        h = list[0]
        if not tails[h]: break
    else:
        raise ValueError("Cannot create a consistent merge order for lists.")
    i = 0
    while i < len(lists):
        list = lists[i]
        if not list:
            lists.remove(list)
            continue
        if list[0] == h:
            list.pop(0)
            if not list:
                lists.remove(list)
                continue
            tails.subtract((list[0],))
        i += 1
    return h

def _merge(lists):
    tails = Counter()
    for list in lists:
        tails.update(list[1:])
    while lists:
        yield _merge_one(lists, tails)

def merge(lists):
    '''Merge several lists, preserving relative order.

    Raises an exception if this is impossible.

    >>> merge([[1,3,5], [2,3,6], [7]])
    [1, 2, 3, 5, 6, 7]
    >>> merge([[], [9,5,1], [10,5,1], [8,5,1], []])
    [9, 10, 8, 5, 1]
    >>> merge([[1, 2], [2, 1]])
    Traceback (most recent call last):
        ...
    ValueError: Cannot create a consistent merge order for lists.
    '''
    return list(_merge([list(l) for l in lists]))

def slice_at(list, value):
    '''Get all items in a list after the first occurrence of value.

    >>> slice_at(['a', 'b', 'c', 'd'], 'b')
    ['c', 'd']

    '''
    return list[list.index(value)+1:]

# SuperMarker is like super() for dispatch targets.
# If dispatcher(val) dispatches with the keys [k1, k2, k3], then
# dispatcher(SuperMarker(k1, val)) will dispatch with the keys [k2, k3].
SuperMarker = namedtuple('SuperMarker', ['key', 'val'])

class _BaseDispatcher(object):
    '''Abstract base class for Dispatcher and DispatchSuper.

    Subclasses must provide .dispatch_mro and ._to_keys(val)
    '''
    def get_default(self):
        for dispatcher in self.dispatch_mro:
            if dispatcher._default:
                return dispatcher._default

    def _dispatch(self, key):
        for dispatcher in self.dispatch_mro:
            v = dispatcher._self_dispatch(key)
            if v is not None:
                return v
        return None

    def dispatch(self, val):
        '''Get a function for a value. Returns None if none is set.'''
        if isinstance(val, SuperMarker):
            start_key, val = val
        else:
            start_key = None
        keys = self._to_keys(val)
        if start_key:
            keys = keys[keys.index(start_key)+1:]
        for key in keys:
            fn = self._dispatch(key)
            if fn:
                return fn
        return self.get_default()

    def raw_call(self, val, *args, **kwargs):
        '''Invoke the dispatch target on val.

        The arguments to the function will be (self, val, *args, **kwargs).
        '''
        fn = self.dispatch(val)
        if fn is None:
            raise NotImplementedError(self._to_keys(val))
        if isinstance(val, SuperMarker):
            val = val[1]
        return fn(self, val, *args, **kwargs)

    def call(self, *args, **kwargs): # pragma: no cover
        # There are no current subclasses that don't override this method
        return self.raw_call(*args, **kwargs)

    def __call__(self, val, *args, **kwargs):
        '''Synonym for self.call()'''
        return self.call(val, *args, **kwargs)


def _resolve_dispatcher(disp_or_disp_wrapper):
    if hasattr(disp_or_disp_wrapper, 'dispatch_mro'):
        return disp_or_disp_wrapper
    return _resolve_dispatcher(disp_or_disp_wrapper.dispatcher)


class Dispatcher(_BaseDispatcher):
    '''Dispatcher that chooses a function based on the first argument.

    A dispatcher is a collection of similar functions plus a mechanism for
    deciding which function to call based on a single value. Each dispatcher has
    a dictionary mapping keys to functions, and a keyfn that maps values to
    lists of keys. It selects a function for a value by invoking the keyfn and
    checking each key until it finds one in its dictionary.

    By default, the keyfn uses the MRO of the type of the first argument:

    >>> d = Dispatcher()
    >>> @d.when(int)
    ... def d_int(d, v):
    ...     return "{} is an integer.".format(v)
    >>> @d.when(bool)
    ... def d_bool(d, v):
    ...     return "{} is a boolean.".format(v)
    >>> d.dispatch(12) is d_int
    True
    >>> d.dispatch(True) is d_bool
    True
    >>> d.dispatch(4+3j) is None
    True


    Dispatchers are callable. When called, a dispatcher will dispatch according
    to the first argument of the call and invoke that function, passing in the
    dispatcher itself and all the call arguments:

    >>> d(12)
    '12 is an integer.'
    >>> d(False)
    'False is a boolean.'
    >>> d(3j)
    Traceback (most recent call last):
        ...
    NotImplementedError: (<class 'complex'>, <class 'object'>)


    You can provide a custom keyfn for more complex dispatch. For example,
    consider the following keyfn that returns all prefix paths of a
    '/'-delimited path:

    >>> import re
    >>> def path_prefixes(s):
    ...     return list(reversed([s[:m.start()] for m in re.finditer('$|/', s)]))
    >>> path_prefixes('foo/bar/baz')
    ['foo/bar/baz', 'foo/bar', 'foo']

    Then a dispatcher with this keyfn will check "foo/bar/baz", "foo/bar", and
    "foo" until it finds a match:

    >>> fmt_error = Dispatcher(keyfn=path_prefixes)
    >>> @fmt_error.when("bad_email")
    ... def general_email(d, v):
    ...     return "Invalid Email"
    >>> @fmt_error.when("bad_email/fmt")
    ... def no_at(d, v):
    ...     return "Invalid Email Format"
    >>> @fmt_error.when("bad_email/fmt/no_at")
    ... def no_at(d, v):
    ...     return "Email must contain an @"
    >>> fmt_error("bad_email")
    'Invalid Email'
    >>> fmt_error("bad_email/fmt/no_at")
    'Email must contain an @'
    >>> fmt_error("bad_email/fmt/bad_domain")
    'Invalid Email Format'


    The key function can be anything that returns a list of keys:

    >>> d = Dispatcher(keyfn=lambda x:range(len(x), 0, -1))
    >>> @d.when(3)
    ... def d3(d, v):
    ...     return "{} is length 3.".format(v)
    >>> @d.when(4)
    ... def d4(d, v):
    ...     return "THERE ARE FOUR LIGHTS!"
    >>> @d.default()
    ... def dn(d, v):
    ...     return "I only know three and four."
    >>> d('foo')
    'foo is length 3.'
    >>> d([1,2,3])
    '[1, 2, 3] is length 3.'
    >>> d({'a', 'b', 'c', 'd'})
    'THERE ARE FOUR LIGHTS!'
    >>> d([])
    'I only know three and four.'
    >>> d('this string has more than four characters')
    'THERE ARE FOUR LIGHTS!'

    If you dispatch on a SuperMarker value, then all keys up to the SuperMarker's
    key will be skipped, similar to the behavior of the builtin super() type:

    >>> fmt_error(SuperMarker('bad_email/fmt', 'bad_email/fmt/no_at'))
    'Invalid Email'

    Finally, each dispatcher has an optional list of parents. You can pass this
    in on initialization, or call the .sub() method of an existing dispatcher to
    create a new dispatcher with the same keyfn and with that dispatcher as its
    sole parent.

    If a dispatcher doesn't have an entry for its key, it will check its parents
    before moving on to the next key. This allows a kind of inheritance:


    >>> my_fmt_error = fmt_error.sub()
    >>> @my_fmt_error.when("bad_email/fmt/bad_domain")
    ... def bad_domain(d, v):
    ...     return "The domain is invalid"
    >>> my_fmt_error("bad_email/fmt/no_at") # calls the parent's dispatch result
    'Email must contain an @'
    >>> my_fmt_error("bad_email/fmt/bad_domain")
    'The domain is invalid'

    Multiple inheritance behaves as you would expect:

    >>> alt_fmt_error = fmt_error.sub()
    >>> @alt_fmt_error.when("bad_email/fmt/no_at")
    ... def alt_no_at(d, v):
    ...     return "You need an @"

    >>> joint_fmt_error = Dispatcher(parents=[my_fmt_error, alt_fmt_error])
    >>> joint_fmt_error("bad_email/fmt/no_at") # inherits no_at from alt_
    'You need an @'
    >>> joint_fmt_error("bad_email/fmt/bad_domain") # inherits bad_domain from my_
    'The domain is invalid'


    Note that the dispatched functions always take the dispatcher itself as the
    first argument. This allows for recursive dispatchers. For example, consider
    dispatch based on types:

    >>> handle = Dispatcher(keyfn=lambda x:type(x).__mro__)
    >>> @handle.when(object)
    ... def handle_obj(handle, obj, indent=''):
    ...     print('{}{}'.format(indent, obj))
    >>> @handle.when(list)
    ... def handle_list(handle, l, indent=''):
    ...     print('{}List:'.format(indent))
    ...     for item in l:
    ...         handle(item, indent+'  ')


    This dispatcher prints objects, splitting apart lists:

    >>> x = [1, 2.0, ['abc', 'foo', 107]]
    >>> handle(x)
    List:
      1
      2.0
      List:
        abc
        foo
        107

    We can "subclass" this handler using the .sub() method, and add custom
    handling for a specific type:

    >>> handle2 = handle.sub()
    >>> @handle2.when(int)
    ... def handle_int(handle, i, indent=''):
    ...     print('{}The number {}'.format(indent, i))

    Because the handle_list function takes the dispatcher itself as its first
    argument, this sub-dispatcher can use the same list handler while ensuring
    that it itself will be used on sub-elements:

    >>> handle2(x)
    List:
      The number 1
      2.0
      List:
        abc
        foo
        The number 107
    '''
    def __init__(self, mapping=None, default=None, keyfn=None, parents=()):
        self.mapping = mapping.copy() if mapping else {}
        self._default = default
        self._default_values = {}
        self._default_factories = {}
        self.keyfn = keyfn
        if parents:
            parents = [_resolve_dispatcher(p) for p in parents]
            self.dispatch_mro = (self,) + tuple(merge(p.dispatch_mro for p in parents))
            if self.keyfn is None:
                self.keyfn = parents[0].keyfn
        else:
            self.dispatch_mro = (self,)

    def _to_keys(self, val):
        '''Convert a value to a list of keys.'''
        if self.keyfn:
            return self.keyfn(val)
        return type(val).__mro__

    def register(self, keys, fn):
        '''Register fn as the target for each key in keys'''
        for key in keys:
            self.mapping[key] = fn
        return fn

    def when(self, *keys):
        '''Decorator version of .register()'''
        def decorate(fn):
            return self.register(keys, fn)
        return decorate

    def set_default(self, fn):
        '''Set the default function to call for keys that aren't defined.

        If None, then invoking a dispatcher on an unrecognized key will raise a
        NotImplementedError.
        '''
        self._default = fn

    def default(self):
        '''Decorator version of set_default'''
        def decorate(fn):
            self.set_default(fn)
            return fn
        return decorate

    def _self_dispatch(self, key):
        if key in self.mapping:
            return self.mapping[key]
        return None

    def sub(self):
        '''Create a new dispatcher with the same keyfn and self as a parent.'''
        return Dispatcher(parents=[self], keyfn=self.keyfn)

    def apply_defaults(self, kwargs):
        for dispatcher in self.dispatch_mro:
            for key, val in dispatcher._default_values.items():
                kwargs.setdefault(key, val)
            for key, val in dispatcher._default_factories.items():
                if key not in kwargs:
                    kwargs[key] = val()
        return kwargs

    def call(self, *args, **kwargs):
        kwargs = self.apply_defaults(kwargs)
        return self.raw_call(*args, **kwargs)

    def default_value(self, key, value):
        self._default_values[key] = value
        self._default_factories.pop(key, None)

    def default_factory(self, key, value_fn):
        self._default_factories[key] = value_fn
        self._default_values.pop(key, None)


class DispatchSuper(_BaseDispatcher):
    '''
    Like super(), but for dispatchers.

    >>> import re
    >>> def path_prefixes(s):
    ...     return list(reversed([s[:m.start()] for m in re.finditer('$|/', s)]))

    >>> disp = Dispatcher(keyfn=path_prefixes)
    >>> @disp.when("recurse")
    ... def disp_recurse(d, s, val):
    ...     print(val)
    ...     if val > 0:
    ...         d(s, val-1)
    >>> disp2 = Dispatcher(parents=[disp])
    >>> @disp2.when("recurse")
    ... def disp2_recurse(d, s, val):
    ...     print("From disp2!")
    ...     DispatchSuper(disp2, d)(s, val)
    >>> disp2("recurse", 5)
    From disp2!
    5
    From disp2!
    4
    From disp2!
    3
    From disp2!
    2
    From disp2!
    1
    From disp2!
    0

    Super supports defaults as well - if the parent has none, it will fail as
    you'd expect:

    >>> @disp2.when("not_implemented_in_parent")
    ... def disp2_niip(d, s, val):
    ...     print("This will hit parent's default, which is not implemented.")
    ...     DispatchSuper(disp2, d)(s, val)
    >>> disp2("not_implemented_in_parent", 5)
    Traceback (most recent call last):
        ...
    NotImplementedError: ['not_implemented_in_parent']

    If a parent dispatcher does have a default, it gets called:

    >>> @disp.default()
    ... def disp_default(d, s, val):
    ...     print("It is implemented now!")

    >>> disp2("not_implemented_in_parent", 5)
    This will hit parent's default, which is not implemented.
    It is implemented now!

    '''
    def __init__(self, above_disp, disp):
        self.disp = _resolve_dispatcher(disp)
        self.above_disp = _resolve_dispatcher(above_disp)
        self.dispatch_mro = slice_at(self.disp.dispatch_mro, self.above_disp)

    def _to_keys(self, val):
        return self.disp._to_keys(val)

    def call(self, val, *args, **kwargs):
        # Need to override base so that we pass in self.disp instead of self
        fn = self.dispatch(val)
        if fn is None:
            raise NotImplementedError(self._to_keys(val))
        return fn(self.disp, val, *args, **kwargs)

def wraps_dispatcher(disp):
    '''
    A decorator for wrapping dispatchers in functions.

    This is especially useful when you want a dispatcher with default values.
    For example, consider a dispatcher that adds a second argument to its first:

    >>> add = Dispatcher()
    >>> @add.when(int)
    ... def add_int(add, value, addition):
    ...     return value + addition

    We can wrap this in a function that defaults additions to 1

    >>> @wraps_dispatcher(add)
    ... def add1(add, value, addition=1):
    ...     return add(value, addition)
    ...

    This behaves just like the original dispatcher:

    >>> add(4, 1)
    5
    >>> add1(4)
    5

    The first argument to the wrapping function is the wrapped dispatcher; this
    makes it possible to proxy .sub() (see below)

    It also proxies the original dispatcher's methods:

    >>> @add1.default()
    ... def add_int_castable(add, value, addition):
    ...     return int(value) + addition
    ...
    >>> add('10', 1)
    11
    >>> add1('12')
    13

    Note that calling `add1.when` actually registered the handler on the
    original `add` dispatcher.

    It also proxies `Dispatcher.sub` correctly:

    >>> add1_sub = add1.sub()
    >>> @add1_sub.when(list)
    ... def add_list_weirdly(add, value, addition):
    ...     return 10 + addition if value else 20 + addition
    ...
    >>> add1_sub([True])
    11
    >>> add1([True])
    Traceback (most recent call last):
      ...
    TypeError: int() argument must be a string or a number, not 'list'

    In the above case, `add1_sub` defines additional behavior for lists, but the
    parent dispatcher is unaffected.

    It's possible to use dispatcher wrappers as explicit parents:

    >>> add_new = Dispatcher(parents=(add1,))

    But in doing so, you're only inheriting the dispatcher structure, not the
    wrapper function:

    >>> add_new(3)
    Traceback (most recent call last):
      ...
    TypeError: add_int() missing 1 required positional argument: 'addition'

    You can get the underlying dispatcher with `.dispatcher` and the underlying
    wrapper function with `.fn`, so if you want to explicitly parent and rewrap
    you could do:

    >>> add_new_disp = Dispatcher(parents=(add1,))
    >>> add_new = wraps_dispatcher(add_new_disp)(add1.fn)
    >>> add_new(3)
    4
    '''
    def wrap(fn):
        @wraps(fn)
        def newfn(*a, **kw):
            return fn(disp, *a, **kw)
        newfn.dispatcher = disp
        newfn.fn = fn
        newfn.get_default = disp.get_default
        newfn.dispatch = disp.dispatch
        newfn.register = disp.register
        newfn.when = disp.when
        newfn.set_default = disp.set_default
        newfn.default = disp.default
        newfn.call = newfn
        def sub(*a, **kw):
            subdisp = disp.sub(*a, **kw)
            return wraps_dispatcher(subdisp)(fn)
        newfn.sub = sub
        return newfn
    return wrap



import vertigo as vg

from .base import Marker, to_typegraph, IGNORE, aggregating_errors
from .base import graphize, validate, dictify, clone, traverse
from .invalid import Invalid

class Tuple(Marker):
    '''The following docstring is ENTIRELY SUSPECT. TODO: Fix it.

    Dictifier for tuples.

    >>> from . import Complex, Int, undictify
    >>> t = Tuple.mkgraph((Int(), Complex()))
    >>> dictify(t, (2, 3j))
    (2, 3j)
    >>> undictify(t, (7, 8+2j))
    (7, (8+2j))
    >>> print(vg.ascii_tree(graphize(t, ("7", "8+2j")), sort=True))
    root: ('7', '8+2j')
      +--0: '7'
      +--1: '8+2j'

    >>> undictify(t, 7j)
    Traceback (most recent call last):
        ...
    Invalid: not_iterable
    >>> undictify(t, (1,))
    Traceback (most recent call last):
        ...
    Invalid: bad_len
    >>> validate(t, (6, 1))
    >>> validate(t, 3)
    Traceback (most recent call last):
        ...
    Invalid: not_iterable
    >>> validate(t, ("this", "tuple", "is", "too", "long"))
    Traceback (most recent call last):
        ...
    Invalid: bad_len
    >>> validate(t, (1.5, "hello"))
    Traceback (most recent call last):
        ...
    Invalid: 0: [type_error], 1: [type_error]

    It does support naming the attributes for the purposes of error reporting
    and walking the tree; see also the NamedTuple dfier.

    >>> t2 = Tuple(field_names=("num", "cnum")).of(num=Int(),cnum=Complex())
    >>> validate(t2, (1.5, "hello"))
    Traceback (most recent call last):
        ...
    Invalid: cnum: [type_error], num: [type_error]

    You must provide either a number of fields or a set of field names:

    >>> Tuple()
    Traceback (most recent call last):
        ...
    ValueError: nfields or field_names is required.
    '''
    def __init__(self, nfields=None, field_names=None):
        super(Tuple, self).__init__()
        if field_names is None:
            if nfields is None:
                raise ValueError("nfields or field_names is required.")
            field_names = [str(i) for i in range(nfields)]
        if nfields is None:
            nfields = len(field_names)
        self.field_names = field_names
        self.nfields = nfields

    def of(self, **kw):
        children = {key:to_typegraph(val) for key, val in kw.items()}
        return vg.PlainGraphNode(self, children)

    @classmethod
    def mkgraph(cls, type_tuple):
        # type_tuple should be a tuple of typegraphs
        children = {str(k):val for k, val in enumerate(type_tuple)}
        return cls(nfields=len(type_tuple)).of(**children)

def apply_tuple(dispgraph, value, kw):
    error_mode = kw.get("error_mode", IGNORE)
    names = dispgraph.marker.field_names
    if error_mode == IGNORE:
        return tuple(dispgraph[n](v, **kw) for n,v in zip(names, value))
    try:
        if len(value) != len(names):
            msg = "Expected iterable of length {}, not {}"
            msg = msg.format(len(names), len(value))
            raise Invalid('bad_len', msg=msg, fatal=True)
    except TypeError:
        msg = "Expected iterable, not {}".format(type(value))
        raise Invalid('not_iterable', msg, fatal=True)
    with aggregating_errors(error_mode) as agg:
        l = []
        for (n, val) in zip(names, value):
            with agg.checking_sub(n):
                l.append(dispgraph[n](val, **kw))
        return tuple(l)


@clone.when(Tuple)
def clone_tuple(dispgraph, value, **kw):
    return apply_tuple(dispgraph, value, kw)

@traverse.when(Tuple)
def traverse_tuple(dispgraph, value, **kw):
    apply_tuple(dispgraph, value, kw)

@graphize.when(Tuple)
def graphize_tuple(dispgraph, value, **kw):
    items = apply_tuple(dispgraph, value, kw)
    names = dispgraph.marker.field_names
    edges = zip(names, items)
    if 'zipval' in dispgraph.extras:
        value = (value, dispgraph.extras.zipval)
    return vg.PlainGraphNode(value, edges)


class NamedTuple(Tuple):
    '''
    Like Tuple, but for namedtuple instances:

    >>> from collections import namedtuple
    >>> from . import Complex, Int, undictify
    >>> WeirdPoint = namedtuple("WeirdPoint", ["num", "cnum"])
    >>> t = NamedTuple(WeirdPoint).of(num=Int(), cnum=Complex())
    >>> p = WeirdPoint(3, 6j)
    >>> dictify(t, p)
    (3, 6j)
    >>> undictify(t, (5, 2+1.0j))
    WeirdPoint(num=5, cnum=(2+1j))
    >>> validate(t, p)
    >>> g = vg.from_dict(dict(num="A Number", cnum="A Complex Number"))
    >>> print(vg.ascii_tree(graphize(t, p, extras_graphs=dict(zipval=g))))
    root: (WeirdPoint(num=3, cnum=6j), None)
      +--num: (3, 'A Number')
      +--cnum: (6j, 'A Complex Number')

    Note that validate explicitly expects the particular namedtuple class, and
    won't settle even for an identical one:

    >>> WeirdPoint2 = namedtuple("WeirdPoint", ["num", "cnum"])
    >>> validate(t, WeirdPoint2(3, 6j))
    Traceback (most recent call last):
        ...
    Invalid: type_error
    '''
    def __init__(self, tuple_type):
        super(NamedTuple, self).__init__(field_names=tuple_type._fields)
        self.tuple_type = tuple_type


@validate.when(NamedTuple)
def validate_namedtuple(dispgraph, value, **kw):
    tuple_type = dispgraph.marker.tuple_type
    if not isinstance(value, tuple_type):
        msg = "Expected {}, got {}".format(tuple_type, type(value))
        raise Invalid('type_error', msg=msg, fatal=True)
    dispgraph.super(NamedTuple)(value, **kw)


@clone.when(NamedTuple)
def clone_namedtuple(dispgraph, value, **kw):
    t = dispgraph.super(NamedTuple)(value, **kw)
    return dispgraph.marker.tuple_type._make(t)


@dictify.when(NamedTuple)
def df_namedtuple(dispgraph, value, **kw):
    '''Explicit dictify to provide a plain tuple.

    Without this, dictify would fall back on the clone() implementation and
    would therefore return a namedtuple of the appropriate type.

    It's ok that undictify falls back on clone, because cloning a plain tuple
    actually will create an appropriate namedtuple.
    '''
    return dispgraph.super(NamedTuple)(value, **kw)



import vertigo as vg

from .cantrips.empty_instance import create_instance

from .base import graphize, validate, dictify, undictify, to_typegraph, traverse
from .base import clone, mutate, aggregating_errors, IGNORE
from .invalid import Invalid
from .schema import Schema, apply_schema

class ObjectMarker(Schema):
    '''Marker for objects that can be assembled by field.

    The Marker is parameterized by a target_cls, which determines what type is
    expected.

    The children of an ObjectMarker node in the typegraph identify the salient
    attributes of the target_cls; dictify will use getattr to extract and store
    those attributes, and undictify will unpack the stored attributes and pass
    them into the marker's .construct function to produce an object with those
    attrs.

    By default, the .construct uses the empty_instance cantrip, which creates
    blank copies of objects and then sets attributes on them in order to bypass
    any side-effects of the class's __init__ function:

    >>> from . import Int
    >>> class Pair(object):
    ...     def __init__(self, x, y):
    ...         print("Creating a pair!")
    ...         self.x, self.y = x, y
    ...     def __repr__(self):
    ...         return 'Pair({!r}, {!r})'.format(self.x, self.y)
    >>> int_pair = ObjectMarker(Pair).of(
    ...     x = Int(),
    ...     y = Int(),
    ... )
    >>> p1 = Pair(2,4)
    Creating a pair!
    >>> p2 = undictify(int_pair, dict(x=2, y=4))
    >>> (p2.x, p2.y) == (p1.x, p1.y)
    True
    >>> undictify(int_pair, (2,4))
    Traceback (most recent call last):
        ...
    Invalid: type_error - Expected a dict, got <type 'tuple'> instead

    You can also give an ObjectMarker a custom construct, which must accept
    keyword arguments of all the typegraph keys.

    >>> int_pair2 = ObjectMarker(Pair, constructor=Pair).of(
    ...     x = Int(),
    ...     y = Int()
    ... )
    >>> p3 = undictify(int_pair2, dict(x=2, y=4))
    Creating a pair!
    >>> (p3.x, p3.y) == (p1.x, p1.y)
    True

    Validate and graphize work as you'd expect:

    >>> validate(int_pair, p3)
    >>> invalid_p = Pair("hi", "ho")
    Creating a pair!
    >>> validate(int_pair, invalid_p)
    Traceback (most recent call last):
        ...
    Invalid: x: [type_error], y: [type_error]
    >>> validate(int_pair, (4,5))
    Traceback (most recent call last):
        ...
    Invalid: type_error - Expected Pair, got tuple

    >>> print(vg.ascii_tree(graphize(int_pair, p3), sort=True))
    root: Pair(2, 4)
      +--x: 2
      +--y: 4
    '''
    target_cls = None

    def construct(self, object_kwargs, **kw):
        if self.constructor:
            return self.constructor(**object_kwargs)
        return create_instance(self.target_cls, object_kwargs)

    def __init__(self, target_cls=None, constructor=None):
        if target_cls is not None:
            self.target_cls = target_cls
        self.constructor = constructor

    def of(self, **kw):
        children = {key:to_typegraph(val) for key, val in kw.items()}
        return vg.PlainGraphNode(self, **children)

def _as_dict(dispgraph, value, default_nones=False):
    result = {}
    for attr in dispgraph.key_iter():
        if hasattr(value, attr):
            result[attr] = getattr(value, attr)
        elif default_nones:
            result[attr] = None
        # If the attr is missing and not default_nones, we don't include it at
        # all in the dict. We assume that later processing will catch this
        # problem.
    return result


def extract_obj(dispgraph, obj, kw, default_nones=False):
    d = _as_dict(dispgraph, obj, default_nones)
    return apply_schema(dispgraph, d, kw, default_nones)


@graphize.when(ObjectMarker)
def graphize_obj(dispgraph, value, **kw):
    d = _as_dict(dispgraph, value)
    g = dispgraph.super(ObjectMarker)(d, **kw)
    g.value = value
    if 'zipval' in dispgraph.extras:
        g.value = (g.value, dispgraph.extras.zipval)
    return g



@mutate.when(ObjectMarker)
def mutate_obj(dispgraph, value, **kw):
    newvals = extract_obj(dispgraph, value, kw)
    for k, v in newvals.items():
        setattr(value, k, v)
    return value


@clone.when(ObjectMarker)
def clone_obj(dispgraph, value, **kw):
    newvals = extract_obj(dispgraph, value, kw)
    return dispgraph.marker.construct(newvals, **kw)


@traverse.when(ObjectMarker)
def traverse_obj(dispgraph, value, **kw):
    extract_obj(dispgraph, value, kw, default_nones=False)


@validate.when(ObjectMarker)
def validate_obj(dispgraph, value, **kw):
    if kw.get('error_mode', IGNORE) != IGNORE:
        marker = dispgraph.marker
        if not isinstance(value, marker.target_cls):
            name = type(value).__name__
            expected = marker.target_cls.__name__
            msg = "Expected {}, got {}".format(expected, name)
            raise Invalid("type_error", msg, fatal=True)
    return dispgraph.parent(validate)(value, **kw)


@dictify.when(ObjectMarker)
def dictify_obj(dispgraph, value, **kw):
    return extract_obj(dispgraph, value, kw, default_nones=True)


@undictify.when(ObjectMarker)
def undictify_obj(dispgraph, value, **kw):
    marker = dispgraph.marker
    with aggregating_errors(kw.get('error_mode', IGNORE)) as agg:
        result = dispgraph.super(ObjectMarker)(value, **kw)
        if agg:
            extra_keys = set(value.keys()) - set(dispgraph.key_iter())
            if extra_keys:
                raise Invalid('unexpected_fields', keys=extra_keys)
    return marker.construct(result, **kw)




import collections

class ImmutableDict(collections.Mapping):
    '''An immutable dictionary.

    For the most part, it behaves like a normal dictionary, except that you
    can't set items:

    >>> repr(ImmutableDict(a=5))
    "ImmutableDict({'a': 5})"
    >>> d = ImmutableDict(dict(a=1, b=2), x=12, y=14)
    >>> assert d == {'a':1, 'b':2, 'x':12, 'y':14}
    >>> assert len(d) == 4
    >>> assert 'a' in d
    >>> assert 'c' not in d
    >>> d['foo'] = 12
    Traceback (most recent call last):
        ...
    TypeError: 'ImmutableDict' object does not support item assignment

    It also has the .overlay method, which behaves like .update() on a normal
    dict, but returns a new ImmutableDict:

    >>> d2 = d.overlay(dict(a=2, foo='bar'))
    >>> d2 == dict(a=2, b=2, x=12, y=14, foo='bar')
    True

    Overlay also provides a strip_none argument, which defaults to False; if it
    is explicitly set to True, then keys in the new dict that are set to None
    will be dropped (instead of replaced by None):

    >>> not_stripped = d.overlay(dict(a=None, b=None))
    >>> not_stripped == dict(a=None, b=None, x=12, y=14)
    True
    >>> stripped = d.overlay(dict(a=None, b=None), strip_none=True)
    >>> stripped == dict(x=12, y=14)
    True

    The .select method creates a subset of this dict:

    >>> subd = d.select(['a', 'x'])
    >>> subd == dict(a=1, x=12)
    True

    The .omit method is like .select, but blacklisting instead of whitelisting:

    >>> subd = d.omit(['y', 'b'])
    >>> subd == dict(a=1, x=12)
    True

    '''
    def __init__(self, *args, **kwargs):
        self.__dict = dict(*args, **kwargs)

    def __getitem__(self, key):
        return self.__dict[key]

    def __contains__(self, key):
        return key in self.__dict

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __repr__(self):
        return "ImmutableDict({!r})".format(self.__dict)

    def overlay(self, other, strip_none=False):
        '''
        Create a copy of one dict with values from a new one.

        >>> d = ImmutableDict(a=1, b=2)
        >>> d2 = d.overlay(dict(a=2, foo='bar'))
        >>> assert d2 == {'a':2, 'b':2, 'foo':'bar'}

        If the strip_none argument is True, keys in the new dict that are None
        will be dropped instead of replaced by None:

        >>> not_stripped = d.overlay(dict(a=None))
        >>> not_stripped == dict(a=None, b=2)
        True
        >>> stripped = d.overlay(dict(a=None), strip_none=True)
        >>> stripped == dict(b=2)
        True
        '''
        d = dict(self)
        d.update(other)
        if strip_none:
            for key, value in other.items():
                if value is None:
                    del d[key]
        return ImmutableDict(d)

    def select(self, keys):
        '''
        Create a subdict of an existing dict.

        >>> d = ImmutableDict(a=1, b=2, c=3, d="four")
        >>> d2 = d.select({'a', 'd'})
        >>> assert d2 == dict(a=1, d="four")
        '''
        return ImmutableDict({k:self[k] for k in keys})

    def omit(self, keys):
        '''
        Create a subdict of an existing dict by omitting keys

        >>> d = ImmutableDict(a=1, b=2, c=3, d="four")
        >>> d2 = d.omit({'a', 'd'})
        >>> assert d2 == dict(b=2, c=3)
        '''
        return ImmutableDict({k:self[k] for k in set(self)-set(keys)})

    def select_overlay(self, *keys, **other):
        '''
        Shortand for self.select(keys).overlay(other).

        >>> d = ImmutableDict(a=1, b=2, c=3, d="four")
        >>> d2 = d.select(['a', 'd']).overlay({'b':5})
        >>> d3 = d.select_overlay('a', 'd', b=5)
        >>> assert d2 == d3
        '''
        return self.select(keys).overlay(other)



import sys
if sys.version >= '3': # pragma: no cover
    unicode = str

from uuid import uuid4

import vertigo as vg

from travesty import SchemaObj, String, Invalid
from travesty import clone, mutate, traverse, graphize, dictify, undictify
from travesty.base import aggregating_errors, IGNORE
from travesty.cantrips import empty_instance
from travesty.schema import apply_schema
from travesty.object_marker import extract_obj


from .docset import DocSet, DoubleLoadException

def make_uid():
    '''Generate a (probably) unique id.'''
    return unicode(uuid4())

class UnloadedDocumentException(Exception):
    '''Raised on attr access of an unloaded document.'''
    def __init__(self, document, *a, **kw):
        self.document = document
        super(UnloadedDocumentException, self).__init__(document, *a, **kw)

class Document(SchemaObj):
    '''Document protocol for non-tree object graphs.

    Every document has a uid, which is a string. Each document is assumed to
    have a tree-like typegraph, but documents can contain references to other
    documents (or even to themselves), and this reference structure not be

    Every document has a uid; subclasses determine what other data they have.

    A Document object can be either loaded or unloaded; the doc.loaded flag
    indicates which a particular one is. A loaded document has data, and behaves
    normally; an unloaded document is a placeholder for a document that is
    stored in some external data store (such as a database) but has not had its
    data loaded.

    An unloaded Document has loaded=False; its uid is set but all the other
    attributes defined by field_types are missing, and trying to access them
    will raise an UnloadedDocumentException.

    In general you will never need to create unloaded documents directly, but
    the undictifier (see below) sometimes will.

    If no uid is provided during initialization, then the Document is given a
    generated uid and has new=True.
    '''
    field_types = dict(
        uid = String(),
    )
    loaded = False

    def __init__(self, uid=None, **attrs):
        self.loaded = True
        # If uid is not specified, generate one randomly
        if uid is None:
            uid = make_uid()
        self.uid = uid
        self._load(**attrs)

    @classmethod
    def _create_unloaded(cls, uid):
        '''Create an unloaded instance of this class.'''
        return empty_instance.create_instance(cls, dict(
            loaded = False,
            uid = uid,
        ))

    def load(self, **attrs):
        if self.loaded:
            raise DoubleLoadException(type(self), self.uid)
        self._load(**attrs)
        return self

    def _load(self, **attrs):
        '''Load an unloaded instance of this class'''
        self.loaded = True
        extras = set(attrs.keys()) - set(self.field_types.keys())
        if extras:
            keylist = ', '.join('"{}"'.format(k) for k in extras)
            msg = "{0}() got an unexpected keyword arguments {1}"
            msg = msg.format(type(self).__name__, keylist)
            raise TypeError(msg)
        for key in self.field_types.keys():
            if key == 'uid': continue
            setattr(self, key, attrs.get(key, None))
        return self

    def __getattr__(self, attr):
        '''Raise UnloadedDocumentException on access to unloaded attributes'''
        if not self.loaded and attr != 'uid' and attr in self.field_types:
            raise UnloadedDocumentException(self)
        return object.__getattribute__(self, attr)

    def _loaded_str(self):
        '''__str__ for loaded instances.

        Subclassess can override this to specify __str__ behavior for loaded
        instances while preserving the default for unloaded instances.
        '''
        return "<{}: {}>".format(type(self).__name__, self.uid)

    def __str__(self):
        if self.loaded:
            return self._loaded_str()
        return "<Unloaded {}: {}>".format(type(self).__name__, self.uid)

    def _loaded_repr(self):
        '''__repr__ for loaded instances.

        Subclassess can override this to specify __repr__ behavior for loaded
        instances while preserving the default for unloaded instances.
        '''
        return "<{}: {}>".format(type(self).__name__, self.uid)

    def __repr__(self):
        if self.loaded:
            return self._loaded_repr()
        return unicode(self)


clone.default_factory("in_docset", lambda: DocSet())

@clone.when(Document.marker_cls)
def clone_document(dispgraph, doc, **kwargs):
    '''Clone a Document.

    This creates an exact duplicate of the document AND all child documents it
    encounters. However, internally it tracks documents it's already cloned, so
    if the same document appears twice it won't be copied twice.

    If a traverse_docs extras graph is provided, then documents won't be cloned
    except where traverse_docs is True.

    Specify new_uids=True if you want clone to generate new uids for cloned
    documents; specify in_docset=<some DocSet> if you want the cloned documents
    to be added to a docset.

    Note that if you specify new_uids=False and pass in a docset that already
    contains some of these documents, those documents will NOT be cloned
    '''
    new_uids = kwargs.get('new_uids', False)
    docset = kwargs['in_docset']
    # If traverse_docs says not to continue, stop here
    if 'traverse_docs' in dispgraph.extras:
        if not dispgraph.extras.traverse_docs:
            return doc
    # Check type here if needed
    if kwargs.get('error_mode', IGNORE) != IGNORE:
        marker = dispgraph.marker
        if not isinstance(doc, marker.target_cls):
            name = type(doc).__name__
            expected = marker.target_cls.__name__
            msg = "Expected {}, got {}".format(expected, name)
            raise Invalid("type_error", msg)
    uid = make_uid() if new_uids else doc.uid
    key = (type(doc), uid)
    if key in docset:
        return docset[key]
    new_doc = docset.create(type(doc), uid)
    if not doc.loaded:
        # Nothing else to copy if the doc is unloaded
        return new_doc
    # Load a dict of the results of all the child calls
    attrs = extract_obj(dispgraph, doc, kwargs)
    new_doc.load(**attrs)
    return new_doc


# Inherits a default in_docset from clone
@undictify.when(Document.marker_cls)
def udf_document(dispgraph, value, **kwargs):
    error_mode = kwargs.get('error_mode', IGNORE)
    # Check type here if needed
    if error_mode != IGNORE:
        if not isinstance(value, dict):
            name = type(value).__name__
            msg = "Expected dict, got {}".format(name)
            raise Invalid("type_error", msg)
        if 'uid' not in value:
            # TODO standardize the invalid naming pattern
            raise Invalid('missing_key:uid', "Document has no uid.")
    uid = value['uid']
    doctype = dispgraph.marker.target_cls
    in_docset = kwargs['in_docset']
    doc = in_docset.get_or_create(doctype, uid)
    # If the input has no keys besides 'uid', and the doctype expects more, then
    # this is an unloaded document and we should just return it
    if (len(value) == 1) and len(doctype.field_types) > 1:
        return doc
    # Otherwise, we need to populate it.
    with aggregating_errors(error_mode):
        attrs = apply_schema(dispgraph, value, kwargs)
        if error_mode != IGNORE:
            # TODO this duplicates logic in doc.load - should maybe combine?
            extra_keys = set(value.keys()) - set(dispgraph.key_iter())
            if extra_keys:
                raise Invalid('unexpected_fields', keys=extra_keys)
    doc.load(**attrs)
    return doc


mutate.default_factory("_tv_docs_processed", lambda: set())

@mutate.when(Document.marker_cls)
def mutate_document(dispgraph, doc, **kwargs):
    '''Mutate a Document.

    This will only mutate a given Document once, even if said document appears
    multiple times during the recursion.

    As with clone(), you can pass `traverse_docs` to explicitly control when
    this will descend into a given document.
    '''
    docs_processed = kwargs['_tv_docs_processed']
    # If we've already mutated this doc, we're done no matter what.
    if doc in docs_processed:
        return doc
    # If traverse_docs says not to enter this doc, we're also done.
    if 'traverse_docs' in dispgraph.extras:
        if not dispgraph.extras.traverse_docs:
            return doc
    docs_processed.add(doc)
    superdisp = dispgraph.super(Document.marker_cls)
    return superdisp(doc, **kwargs)


dictify.default_factory('_tv_docs_processed', lambda: set())

@dictify.when(Document.marker_cls)
def dictify_document(dispgraph, doc, **kwargs):
    '''Dictify for Document.

    This has the same behavior as clone, except that where clone would create
    a new object, this instead returns a complete serialized object, and where
    clone would return the original object, this instead returns
    dict(uid=doc.uid)
    '''
    docs_processed = kwargs['_tv_docs_processed']
    # If we've already done this doc, just return a stub
    if doc in docs_processed:
        return dict(uid=doc.uid)
    # If traverse_docs says not to enter this doc, just return a stub
    if 'traverse_docs' in dispgraph.extras:
        if not dispgraph.extras.traverse_docs:
            return dict(uid=doc.uid)
    docs_processed.add(doc)
    superdisp = dispgraph.super(Document.marker_cls)
    return superdisp(doc, **kwargs)


graphize.default_factory("_tv_docs_cache", lambda: {})

@graphize.when(Document.marker_cls)
def graphize_document(dispgraph, doc, **kwargs):
    cache = kwargs['_tv_docs_cache']
    if doc in cache:
        # Already done this one
        return cache[doc]
    # Restrict to uid if the doc isn't loaded OR we're not supposed to traverse.
    superdisp = dispgraph.super(Document.marker_cls)
    if not doc.loaded:
        superdisp = superdisp.restrict(['uid'])
    elif 'traverse_docs' in dispgraph.extras:
        if not dispgraph.extras.traverse_docs:
            superdisp = superdisp.restrict(['uid'])
    cache[doc] = vg.PlainGraphNode()
    new = superdisp(doc, **kwargs)
    cache[doc].value = new.value
    cache[doc]._edges = new._edges
    return cache[doc]


traverse.default_factory("_tv_docs_processed", lambda: set())

@traverse.when(Document.marker_cls)
def traverse_document(dispgraph, doc, **kwargs):
    docs_processed = kwargs['_tv_docs_processed']
    # If we've already done this doc, there's nothing to d
    if doc in docs_processed:
        return
    # If we don't traverse here, then we're done
    if 'traverse_docs' in dispgraph.extras:
        if not dispgraph.extras.traverse_docs:
            return
    docs_processed.add(doc)
    superdisp = dispgraph.super(Document.marker_cls)
    # We use getattr so that if you pass a non-document in we won't break until
    # superdisp (which should already handle exceptions)
    if not getattr(doc, 'loaded', False):
        # Unloaded doc: only traverse uid
        return superdisp.restrict(['uid'])(doc, **kwargs)
    # traverse normally
    return superdisp(doc, **kwargs)



import vertigo as vg

from .base import Marker, Traversable, to_typegraph, IGNORE
from .base import graphize, validate, clone, dictify, undictify
from .invalid import Invalid


class UnknownType(Exception):
    '''Raised when a Polymorph encounters an unknown type.'''
    pass


class Polymorph(Marker):
    '''
    A marker for selecting between several types.

    Generally you want to assemble these using .mkgraph, which takes a mapping
    from string names to either traversable types or (types, typegraph) pairs.

    When operating on an object, the Polymorph will select which of its child
    typegraphs to use based on the type of the object it encounters; when
    dictifying, it will store the string name so that later undictification can
    determine what type was here.

    For example, consider a simple Point type:

    >>> import travesty as tv
    >>> class Point(tv.SchemaObj):
    ...     field_types = dict(x=tv.Number(), y=tv.Complex())
    ...     def __repr__(self): return "Point({}, {})".format(self.x, self.y)
    ...

    We can create a typegraph representing a value that could be a Point or a
    plain number as follows:

    >>> NumOrPoint = Polymorph.mkgraph({
    ...     'point':Point,
    ...     'num':((int, float), tv.Number()),
    ... })

    This typegraph indicates that values of type Point should be processed using
    Point's typegraph, and values of type int or float should be processed using
    the typegraph Number() (well, to_typegraph(Number())). This means that
    various dispatchers can be used on both types:

    >>> tv.clone(NumOrPoint, 12)
    12
    >>> tv.clone(NumOrPoint, Point(x=1, y=2))
    Point(1, 2)
    >>> tv.validate(NumOrPoint, 12)
    >>> tv.validate(NumOrPoint, Point(x=1, y=2))

    Validate and other error-checking will fail on unknown types:

    >>> tv.validate(NumOrPoint, "blah")
    Traceback (most recent call last):
    ...
    Invalid: type_error - Unrecognized type: <type 'str'>
    >>> tv.clone(NumOrPoint, "blah")
    Traceback (most recent call last):
    ...
    UnknownType: str
    >>> tv.clone(NumOrPoint, "blah", error_mode=tv.CHECK)
    Traceback (most recent call last):
    ...
    Invalid: type_error: Unrecognized type: <type 'str'>

    Dictify returns a tuple of (type id, dictified value), so that you can later
    determine what type you had:

    >>> tv.dictify(NumOrPoint, 12)
    ('num', 12)
    >>> dictify(NumOrPoint, 3.5)
    ('num', 3.5)
    >>> dictify(NumOrPoint, Point(x=1, y=3)) == ('point', {'x':1, 'y':3})
    True

    Undictify expects such a tuple:

    >>> tv.undictify(NumOrPoint, ('num', 42))
    42
    >>> p = tv.undictify(NumOrPoint, ('point', {'x':-1, 'y':2+2j}))
    >>> p.x
    -1
    >>> p.y
    (2+2j)

    The input to undictify MUST be a list or tuple of length 2, and the first
    item must be known to the Polymorph:

    >>> tv.undictify(NumOrPoint, 42)
    Traceback (most recent call last):
        ...
    Invalid: type_error
    >>> tv.undictify(NumOrPoint, ('num', 42, 'something extra for no reason'))
    Traceback (most recent call last):
        ...
    Invalid: bad_list
    >>> tv.undictify(NumOrPoint, ('weird_key', 42))
    Traceback (most recent call last):
        ...
    Invalid: bad_typename - weird_key
    >>> tv.undictify(NumOrPoint, ('weird_key', 42), error_mode=IGNORE)
    Traceback (most recent call last):
        ...
    KeyError: weird_key

    Traversal with zipvals expects the graph to have keys matching the main
    graph structure, as always, but note that the graph value at the actual
    Polymorph will be ignored:

    >>> g = vg.from_dict(dict(_self="Ignored", num="A number", point=dict(
    ... _self="A Point", x="The X", y="The Y")))
    >>> print(vg.ascii_tree(graphize(NumOrPoint, 3, extras_graphs=dict(zipval=g))))
    root: (3, 'A number')
    >>> print(vg.ascii_tree(graphize(NumOrPoint, p, extras_graphs=dict(zipval=g)), sort=True))
    root: (Point(-1, (2+2j)), 'A Point')
      +--x: (-1, 'The X')
      +--y: ((2+2j), 'The Y')

    Note that the initialization above is short for the longer form:
    >>> NumOrPoint = Polymorph({'point':Point, 'num':(int, float)}).of(
    ...     point = Point, num=tv.Number()
    ... )


    '''
    def __init__(self, mapping):
        super(Polymorph, self).__init__()
        self.cls_to_name = {}
        for name, classes in mapping.items():
            if not isinstance(classes, (list, tuple)):
                classes = (classes,)
            for cls in classes:
                self.cls_to_name[cls] = name

    def name_for_type(self, typ):
        '''Given a type, get the corresponding name.'''
        for basetype in typ.__mro__:
            if basetype in self.cls_to_name:
                return self.cls_to_name[basetype]
        raise UnknownType(typ.__name__)

    def name_for_val(self, value):
        '''Given an object, get the corresponding name.'''
        return self.name_for_type(type(value))

    def of(self, **kw):
        children = {key:to_typegraph(val) for key, val in kw.items()}
        return vg.PlainGraphNode(self, **children)

    @classmethod
    def mkgraph(cls, mapping):
        # Mapping should be name:(type_or_types, typegraph) or name:Traversable
        lookup = {}
        children = {}
        for name, data in mapping.items():
            if isinstance(data, type) and issubclass(data, Traversable):
                lookup[name] = (data,)
                children[name] = data.typegraph
            else:
                lookup[name], children[name] = data
        return cls(lookup).of(**children)


def apply_pmorph(dispgraph, value, error_mode=IGNORE, **kw):
    '''Returns (name, dictified_value), where name is the polymorphic id.'''
    kw['error_mode'] = error_mode
    try:
        name = dispgraph.marker.name_for_val(value)
    except UnknownType:
        if error_mode == IGNORE:
            raise
        raise Invalid("type_error", "Unrecognized type: {}".format(type(value)))
    value = dispgraph[name](value, **kw)
    return (name, value)


@clone.when(Polymorph)
def clone_pmorph(dispgraph, value, error_mode=IGNORE, **kw):
    name, value = apply_pmorph(dispgraph, value, error_mode, **kw)
    return value


@dictify.when(Polymorph)
def dictify_pmorph(dispgraph, value, error_mode=IGNORE, **kw):
    '''Returns (name, dictified_value), where name is the polymorphic id.'''
    return apply_pmorph(dispgraph, value, error_mode, **kw)


@undictify.when(Polymorph)
def undictify_pmorph(dispgraph, value, error_mode=IGNORE, **kw):
    kw['error_mode'] = error_mode
    if error_mode != IGNORE:
        if not isinstance(value, (list, tuple)):
            raise Invalid('type_error')
        if len(value) != 2:
            raise Invalid('bad_list')
    name, value = value
    if error_mode != IGNORE and name not in dispgraph:
        raise Invalid('bad_typename', name)
    return dispgraph[name](value, **kw)


@validate.when(Polymorph)
def validate_pmorph(dispgraph, value, **kw):
    apply_pmorph(dispgraph, value, **kw)


@graphize.when(Polymorph)
def graphize_pmorph(dispgraph, value, **kw):
    name = dispgraph.marker.name_for_val(value)
    return dispgraph[name](value, **kw)



from collections import OrderedDict

import vertigo as vg

from .invalid import Invalid
from .base import Marker, graphize, traverse, mutate, clone
from .base import to_typegraph, aggregating_errors, IGNORE


class Schema(Marker):
    '''Abstract Marker for schema-structured objects.'''

    def of(self, **kwargs):
        children = {key:to_typegraph(val) for key, val in kwargs.items()}
        return vg.PlainGraphNode(self, **children)


def apply_schema(dispgraph, value, kw, default_nones=True):
    '''Apply a dispgraph to each element in value.

    This also handles error checking - if agg is not None, this will typecheck
    value and recurse to each element within agg.checking_sub().
    '''
    error_mode = kw.get('error_mode', IGNORE)
    def get(key):
        if default_nones:
            return value.get(key, None)
        return value[key]
    result = OrderedDict()
    if error_mode == IGNORE:
        for (key, subgraph) in dispgraph.edge_iter():
            result[key] = subgraph(get(key), **kw)
        return result
    with aggregating_errors(error_mode) as agg:
        if not isinstance(value, dict):
            msg = 'Expected a dict, got {} instead'.format(type(value))
            raise Invalid("type_error", msg, fatal=True)
        for key, subgraph in dispgraph.edge_iter():
            with agg.checking_sub(key):
                if key not in value and not default_nones:
                    raise Invalid("missing_attr")
                val = value.get(key, None)
                result[key] = subgraph(val, **kw)
        return result


@graphize.when(Schema)
def graphize_schema(dispgraph, value, **kw):
    edges = apply_schema(dispgraph, value, kw).items()
    if 'zipval' in dispgraph.extras:
        value = (value, dispgraph.extras.zipval)
    return vg.PlainGraphNode(value, edges)


@traverse.when(Schema)
def traverse_schema(dispgraph, value, **kw):
    apply_schema(dispgraph, value, kw, default_nones=False)


@clone.when(Schema)
def clone_schema(dispgraph, value, **kw):
    return apply_schema(dispgraph, value, kw)


@mutate.when(Schema)
def mutate_schema(dispgraph, value, **kw):
    error_mode = kw.get('error_mode', IGNORE)
    if error_mode != IGNORE and not isinstance(value, dict):
        msg = 'Expected a dict, got {} instead'.format(type(value))
        raise Invalid("type_error", msg, fatal=True)
    value.update(apply_schema(dispgraph, value, kw))
    return value



import vertigo as vg

from . import Wrapper, graphize, clone, traverse

class Optional(Wrapper):
    '''Wrapper that indicates the value could be None.

    >>> import travesty as tv

    None fails validation, won't always undictify, etc.

    >>> tv.validate(tv.Int(), None)
    Traceback (most recent call last):
        ...
    Invalid: type_error
    >>> tv.undictify(tv.List().of(tv.Int()), None)
    Traceback (most recent call last):
        ...
    Invalid: type_error

    Wrapping it in Optional fixes this:

    >>> tv.validate(Optional.wrap(tv.Int()), None)
    >>> tv.undictify(Optional.wrap(tv.Int()), None) is None
    True

    '''
    pass

@graphize.when(Optional)
def graphize_optional(dispgraph, value, **kw):
    if value is None:
        if 'zipval' in dispgraph.extras:
            return vg.PlainGraphNode((None, dispgraph.extras.zipval))
        return vg.PlainGraphNode(None)
    opt = dispgraph.marker
    return dispgraph.for_marker(opt.marker)(value, **kw)


@clone.when(Optional)
def clone_optional(dispgraph, value, **kw):
    if value is None:
        return None
    opt = dispgraph.marker
    return dispgraph.for_marker(opt.marker)(value, **kw)


@traverse.when(Optional)
def traverse_optional(dispgraph, value, **kw):
    if value is None:
        return
    opt = dispgraph.marker
    dispgraph.for_marker(opt.marker)(value, **kw)



from .base import Leaf

class Passthrough(Leaf):
    '''An explicitly ignored leaf value.

    All the various travesty-defined dispatchers will pass values through
    unchanged:

    >>> import travesty as tv
    >>> tv.dictify(Passthrough(), 12)
    12
    >>> tv.undictify(Passthrough(), {'x': [1, 2, 3]})
    {'x': [1, 2, 3]}
    >>> tv.validate(Passthrough(), Exception("This could be any object"))
    >>>
    '''
    # No code needed - Leaf does this already. Passthrough is just here to be
    # more explicit.



import sys
from collections import OrderedDict
if sys.version >= '3': # pragma: no cover
    basestring = str

import vertigo as vg

from .invalid import Invalid
from .base import graphize, traverse, clone, mutate, validate
from .base import Marker, IGNORE, to_typegraph, aggregating_errors
from .schema import Schema

class SchemaMapping(Schema):
    '''Marker for structured dicts.

    The children of this node in the typegraph indicate the types of the keyed
    attributes to traverse.

    >>> from datetime import date
    >>> from . import Int, Date, undictify, dictify
    >>> G = vg.PlainGraphNode

    >>> bday = date(1985, 9, 16)
    >>> schema = G(SchemaMapping(), [("age",G(Int())), ("birthday",G(Date()))])
    >>> def mk(age, birthday):
    ...     return OrderedDict([('age', age), ('birthday', birthday)])
    >>> print(vg.ascii_tree(graphize(schema, mk(age=27, birthday=bday))))
    root: OrderedDict([('age', 27), ('birthday', datetime.date(1985, 9, 16))])
      +--age: 27
      +--birthday: datetime.date(1985, 9, 16)

    >>> validate(schema, mk(age=27, birthday=bday))
    >>> validate(schema, mk(age=27.1, birthday="not a date"))
    Traceback (most recent call last):
    ...
    Invalid: age: [type_error], birthday: [type_error]

    >>> cstruct = dictify(schema, mk(age=27, birthday=bday))
    >>> cstruct == mk(age=27, birthday=u'1985-09-16')
    True
    >>> undictify(schema, cstruct) == mk(age=27, birthday=bday)
    True

    The 'extra_field_policy' flag should be 'save', 'discard', or 'error'. This
    controls what happens on undictifiance when unexpected fields are present.

    'save': un/dictify preserves extra keys; validate ignores them

    >>> schema = G.build(dict(_self=SchemaMapping('save'), age=Int()))
    >>> v = undictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27, 'color': 'blue'}
    True
    >>> validate(schema, v)
    >>> v = dictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27, 'color': 'blue'}
    True

    'discard': un/dictify discards extra keys; validate complains about them

    >>> schema = G.build(dict(_self=SchemaMapping('discard'), age=Int()))
    >>> v = undictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27}
    True
    >>> validate(schema, v)
    >>> validate(schema, {'age':27, 'color':'blue'})
    Traceback (most recent call last):
    ...
    Invalid: unexpected_fields - {'keys': set(['color'])}
    >>> v = dictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27}
    True

    'error': undictify and validate complain; dictify discards

    >>> schema = G.build(dict(_self=SchemaMapping('error'), age=Int()))
    >>> v = undictify(schema, {'age':27, 'color':'blue'})
    Traceback (most recent call last):
    ...
    Invalid: unexpected_fields - {'keys': set(['color'])}
    >>> validate(schema, {'age':27, 'color':'blue'})
    Traceback (most recent call last):
    ...
    Invalid: unexpected_fields - {'keys': set(['color'])}
    >>> v = dictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27}
    True

    >>> validate(schema, 12)
    Traceback (most recent call last):
    ...
    Invalid: type_error - Expected dict, got <type 'int'>
    '''
    def __init__(self, extra_field_policy="discard"):
        super(SchemaMapping, self).__init__()
        self.extra_field_policy=extra_field_policy


@validate.when(SchemaMapping)
def validate_mapping(dispgraph, value, **kw):
    error_mode = kw.get('error_mode', IGNORE)
    marker = dispgraph.marker
    with aggregating_errors(error_mode) as agg:
        dispgraph.super(SchemaMapping)(value, **kw)
        if agg and marker.extra_field_policy in ['discard', 'error']:
            extra_keys = set(value.keys()) - set(dispgraph.key_iter())
            if extra_keys:
                agg.own_error(Invalid('unexpected_fields', keys=extra_keys))


@clone.when(SchemaMapping)
def clone_mapping(dispgraph, value, **kw):
    error_mode = kw.get('error_mode', IGNORE)
    marker = dispgraph.marker
    with aggregating_errors(error_mode) as agg:
        result = dispgraph.super(SchemaMapping)(value, **kw)
        extra_keys = set(value.keys()) - set(dispgraph.key_iter())
        if extra_keys:
            if agg and marker.extra_field_policy == 'error':
                raise Invalid('unexpected_fields', keys=extra_keys)
            if marker.extra_field_policy == 'save':
                for key in extra_keys:
                    result[key] = value[key]
    return result


@mutate.when(SchemaMapping)
def mutate_mapping(dispgraph, value, **kw):
    newval = clone_mapping(dispgraph, value, **kw)
    value.update(newval)
    return value


class StrMapping(Marker):
    '''Marker for dicts with string keys and homogenous values.

    The sub child in the typegraph determines the types of the mapping's values.

    For example, here's a type that maps strings to dates:

    >>> from datetime import date, timedelta
    >>> from collections import OrderedDict
    >>> import travesty as tv
    >>> DateMap = StrMapping().of(tv.Date())
    >>> d1, d2 = date(1985, 9, 16), date(1980, 3, 17)
    >>> example = OrderedDict([
    ...     ("foo", d1),
    ...     ("bar", d2)])
    >>> print(vg.ascii_tree(graphize(DateMap, example), sort=True))
    root: OrderedDict([('foo', datetime.date(1985, 9, 16)), ('bar', datetime.date(1980, 3, 17))])
      +--bar: datetime.date(1980, 3, 17)
      +--foo: datetime.date(1985, 9, 16)

    Dictifying it will map each key to its dictified value:

    >>> cstruct = tv.dictify(DateMap, example)
    >>> cstruct == {'foo':'1985-09-16', 'bar':'1980-03-17'}
    True
    >>> tv.undictify(DateMap, cstruct) == dict(example)
    True

    The validators expect a dict, and will complain if non-string keys are
    present:

    >>> tv.undictify(DateMap, None)
    Traceback (most recent call last):
    ...
    Invalid: type_error - Expected dict, got <type 'Nonetype'>
    >>> tv.undictify(DateMap, {12:d1})
    Traceback (most recent call last):
    ...
    Invalid: value_error/bad_keys - Bad keys - {'keys': [12]}
    >>> tv.validate(DateMap, example)
    >>> tv.validate(DateMap, {'foo':'not a date'})
    Traceback (most recent call last):
    ...
    Invalid: foo: [type_error]
    >>> tv.validate(DateMap, {12:d1})
    Traceback (most recent call last):
    ...
    Invalid: value_error/bad_keys - Bad keys - {'keys': [12]}

    Mutate will leave the keys unchanged:

    >>> Str2Str = StrMapping().of(tv.String())
    >>> double = tv.mutate.sub()
    >>> @double.when(tv.String)
    ... def double_string(dispgraph, value, **kw): return value*2
    ...
    >>> x = dict(a='b', c='d')
    >>> double(Str2Str, x) == dict(a='bb', c='dd')
    True
    '''
    def of(self, sub):
        return vg.PlainGraphNode(self, sub = to_typegraph(sub))

def apply_strmap(dispgraph, value, kw):
    '''Apply a dispgraph to each element in value.

    This also handles error checking - if agg is not None, this will typecheck
    value and recurse to each element within agg.checking_sub().
    '''
    error_mode = kw.get('error_mode', IGNORE)
    vfn = lambda x: dispgraph['sub'](x, **kw)
    if error_mode == IGNORE:
        return {key:vfn(val) for (key, val) in value.items()}
    if not isinstance(value, dict):
        msg = "Expected dict, got {}".format(type(value))
        raise Invalid("type_error", msg, fatal=True)
    result = OrderedDict()
    bad_keys = []
    with aggregating_errors(error_mode) as agg:
        for key, val in value.items():
            if not isinstance(key, basestring):
                bad_keys.append(key)
                continue
            with agg.checking_sub(key):
                result[key] = vfn(val)
        if bad_keys:
            raise Invalid("value_error/bad_keys", "Bad keys", keys=bad_keys)
    return result


@graphize.when(StrMapping)
def graphize_strmap(dispgraph, value, **kw):
    edges = apply_strmap(dispgraph, value, kw).items()
    if 'zipval' in dispgraph.extras:
        value = (value, dispgraph.extras.zipval)
    return vg.PlainGraphNode(value, edges)


@clone.when(StrMapping)
def clone_strmap(dispgraph, value, **kw):
    return apply_strmap(dispgraph, value, kw)


@mutate.when(StrMapping)
def mutate_strmap(dispgraph, value, **kw):
    value.update(apply_strmap(dispgraph, value, kw))
    return value


@traverse.when(StrMapping)
def traverse_strmap(dispgraph, value, **kw):
    apply_strmap(dispgraph, value, kw)


class UniMapping(Marker):
    '''Marker for dicts with homogenous keys and homogenous values.

    The key and val children in the typegraph determine the types of the
    mapping's keys and values, respectively.

    >>> from collections import OrderedDict
    >>> import travesty as tv
    >>> from datetime import date, timedelta
    >>> DateToNumList = UniMapping().of(
    ...     key=tv.Date(),
    ...     val=tv.List().of(tv.Int()),
    ... )
    >>> d1, d2 = date(1985, 9, 16), date(1980, 3, 17)
    >>> example = OrderedDict([(d1,[1,2,3,4]), (d2,[6,12,-4])])

    Graphize will label the keys `key_i` and the values `value_i`:

    >>> print(vg.ascii_tree(tv.graphize(DateToNumList, example), sort=True))
    root: OrderedDict([(datetime.date(1985, 9, 16), [1, 2, 3, 4]), (datetime.date(1980, 3, 17), [6, 12, -4])])
      +--key_0: datetime.date(1985, 9, 16)
      +--key_1: datetime.date(1980, 3, 17)
      +--value_0: [1, 2, 3, 4]
      |  +--0: 1
      |  +--1: 2
      |  +--2: 3
      |  +--3: 4
      +--value_1: [6, 12, -4]
         +--0: 6
         +--1: 12
         +--2: -4

    Dictifying will map each dictified key to its dictified value:

    >>> cstruct = tv.dictify(DateToNumList, example)
    >>> cstruct == {u'1985-09-16':[1,2,3,4], '1980-03-17':[6,12,-4]}
    True
    >>> tv.undictify(DateToNumList, cstruct) == example
    True

    Validation expects a dictionary:

    >>> tv.validate(DateToNumList, example)
    >>> tv.validate(DateToNumList, 12)
    Traceback (most recent call last):
    ...
    Invalid: type_error - Expected dict, got <type 'int'>

    Exceptions will also use `key_i`/`value_i` to indicate sub-problems:

    >>> try:
    ...     tv.undictify(DateToNumList, OrderedDict([
    ...         ('1985-09-16', None),
    ...         ('hello', [1, 'hi']),
    ...     ]))
    ... except Invalid as e:
    ...     print(vg.ascii_tree(e.as_graph(), sort=True))
    root: []
      +--key_1: [SingleInvalid('bad_format',)]
      +--value_0: [SingleInvalid('type_error',)]

    Caveat: Note that if the value passed in is not an OrderedDict, then there's
    no guarantee of the numerical order of the errors or of the graphize result.
    There's not really a good way to handle this; use OrderedDict if you need
    to be able to map problems back to their specific keys.

    Mutate will replace keys if they change:

    >>> one_day_more = tv.mutate.sub()
    >>> @one_day_more.when(tv.Date)
    ... def tomorrow(dispgraph, value, **kw):
    ...     return value + timedelta(days=1)
    ...
    >>> _ = one_day_more(DateToNumList, example)
    >>> example[date(1985, 9, 17)]
    [1, 2, 3, 4]

    '''
    def of(self, key, val):
        key, val = to_typegraph(key), to_typegraph(val)
        return vg.PlainGraphNode(self, key=key, val=val)


def apply_unimap(dispgraph, value, kw):
    '''Apply a dispgraph to each element in value.

    This also handles error checking - if agg is not None, this will typecheck
    value and recurse to each element within agg.checking_sub().
    '''
    error_mode = kw.get('error_mode', IGNORE)
    kfn = lambda x: dispgraph['key'](x, **kw)
    vfn = lambda x: dispgraph['val'](x, **kw)
    result = OrderedDict()
    if error_mode == IGNORE:
        for (key, val) in value.items():
            result[kfn(key)] = vfn(val)
        return result
    if not isinstance(value, dict):
        msg = "Expected dict, got {}".format(type(value))
        raise Invalid("type_error", msg, fatal=True)
    with aggregating_errors(error_mode) as agg:
        for i, (key, val) in enumerate(value.items()):
            with agg.checking_sub('key_{}'.format(i)):
                key = kfn(key)
            with agg.checking_sub('value_{}'.format(i)):
                val = vfn(val)
            result[key] = val
    return result

@graphize.when(UniMapping)
def graphize_unimap(dispgraph, value, **kw):
    result = apply_unimap(dispgraph, value, kw)
    edges = []
    for i, (key, val) in enumerate(result.items()):
        edges.append(('key_{}'.format(i), key))
        edges.append(('value_{}'.format(i), val))
    if 'zipval' in dispgraph.extras:
        value = (value, dispgraph.extras.zipval)
    return vg.PlainGraphNode(value, edges)


@clone.when(UniMapping)
def clone_unimap(dispgraph, value, **kw):
    return apply_unimap(dispgraph, value, kw)


@mutate.when(UniMapping)
def mutate_unimap(dispgraph, value, **kw):
    new_values = apply_unimap(dispgraph, value, kw)
    # Clear in case the keys were mutated
    value.clear()
    value.update(new_values)
    return value


@traverse.when(UniMapping)
def traverse_unimap(dispgraph, value, **kw):
    apply_unimap(dispgraph, value, kw)




'''

>>> from travesty import String, Optional
>>> from travesty import dictify, undictify, traverse, validate, graphize
>>> import vertigo as vg
>>> class Node(Document):
...     # field_types is a fn because python lacks recursive class definitions
...     field_types = lambda cls: dict(
...         name=String(),
...         next=Optional.wrap(cls)
...     )
...     def __init__(self, uid, name, next=None):
...         super(Node, self).__init__(uid)
...         self.name, self.next = name, next
...     def _loaded_repr(self):
...         return "Node({!r}, {!r})".format(self.uid, self.name)
>>> Node._finalize_typegraph()

You can create two nodes that point to each other:

>>> n1 = Node(u"node1", u"First")
>>> n2 = Node(u"node2", u"Second", next=n1)
>>> n1.next = n2

Dictify will only serialize any given document once - when there are multiple
occurences of the same document, all the other occurrences will serialize just
to {'uid':doc.uid}. This means you can dictify recursive objects:

>>> s = dictify(Node, n1)
>>> s == {
... 'uid': 'node1',
... 'name': 'First',
... 'next': {
...     'uid': 'node2',
...     'name': 'Second',
...     'next': {'uid': 'node1'}
... }}
True


>>> validate(Node, n1)
>>> n2.next = None
>>> validate(Node, n1) # but now it won't.
>>> print(vg.ascii_tree(graphize(Node, n1), sort=True))
root: Node(u'node1', u'First')
  +--name: u'First'
  +--next: Node(u'node2', u'Second')
  |  +--name: u'Second'
  |  +--next: None
  |  +--uid: u'node2'
  +--uid: u'node1'



You can also create a DocSet from existing documents, but the DocSet will
NOT automatically search for other referenced documents:

>>> docset = DocSet([n1])
>>> docset[Node, 'node1'] is n1
True
>>> (Node, 'node2') in docset
False
>>> docset.get(Node, 'node2') is None
True

You can add other documents explicitly by calling add:

>>> docset.add(n2)
>>> docset[Node, 'node2'] is n2
True

This will fail if a document with that id is already in the set:

>>> docset.add(n2)
Traceback (most recent call last):
    ...
ValueError: Duplicate uid node2 for type <class 'travesty.document.Node'>

Finally, the explicit DocSet.create(type, uid) function creates an unloaded
document, or fails if that document already exists:

>>> docset.create(Node, 'node3')
<Unloaded Node: node3>
>>> docset.create(Node, 'node1')
Traceback (most recent call last):
    ...
ValueError: Duplicate uid node1 for type <class 'travesty.document.Node'>
'''


from .document import Document, UnloadedDocumentException, DoubleLoadException
from .docset import DocSet

__all__ = ['Document', 'DocSet', 'UnloadedDocumentException',
    'DoubleLoadException']




import vertigo as vg

from .base import Marker, graphize, traverse, clone, mutate
from .base import to_typegraph, aggregating_errors, IGNORE
from .invalid import Invalid

class List(Marker):
    '''Marker for homogenous lists.

    The 'sub' child in the typegraph indicates the marker for all items in the
    list.

    For example:

    >>> from travesty import Int, Optional, dictify, undictify, validate
    >>> l = List().of(Int())

    >>> print(vg.ascii_tree(graphize(l, [1, 2, 3]), sort=True))
    root: [1, 2, 3]
      +--0: 1
      +--1: 2
      +--2: 3

    >>> undictify(l, [1, 4, 5])
    [1, 4, 5]
    >>> dictify(l, [1, 2, 3])
    [1, 2, 3]

    >>> validate(l, [1, 2, 3])
    >>> try:
    ...     validate(l, [1, None, 'hi'])
    ... except Invalid as e:
    ...     print(vg.ascii_tree(e.as_graph(), sort=True))
    root: []
      +--1: [SingleInvalid('type_error',)]
      +--2: [SingleInvalid('type_error',)]
    >>> l = List().of(Optional.wrap(Int()))
    >>> validate(l, [1, None, 'hi'])
    Traceback (most recent call last):
    ...
    Invalid: 2: [type_error]
    >>> validate(l, 12)
    Traceback (most recent call last):
    ...
    Invalid: type_error
    '''
    def of(self, sub):
        return vg.PlainGraphNode(self, sub=to_typegraph(sub))


def apply_list(dispgraph, value, kw):
    '''Apply a dispgraph to each element in value.

    This also handles error checking - if agg is not None, this will typecheck
    value and recurse to each element within agg.checking_sub().
    '''
    error_mode = kw.get('error_mode', IGNORE)
    if error_mode == IGNORE:
        return [dispgraph['sub'](v, **kw) for v in value]
    with aggregating_errors(error_mode) as agg:
        if not isinstance(value, (list, tuple)):
            msg = "Expected list, got {}".format(type(value).__name__)
            raise Invalid("type_error", msg, fatal=True)
        result = []
        for i, v in enumerate(value):
            with agg.checking_sub(str(i)):
                result.append(dispgraph['sub'](v, **kw))
        return result


@graphize.when(List)
def graphize_list(dispgraph, value, **kw):
    edges = apply_list(dispgraph, value, kw)
    edges = ((str(i), v) for (i,v) in enumerate(edges))
    if 'zipval' in dispgraph.extras:
        value = (value, dispgraph.extras.zipval)
    return vg.PlainGraphNode(value, edges)


@traverse.when(List)
def traverse_list(dispgraph, value, **kw):
    apply_list(dispgraph, value, kw)


@mutate.when(List)
def mutate_list(dispgraph, value, **kw):
    value[:] = apply_list(dispgraph, value, kw)
    return value


@clone.when(List)
def clone_list(dispgraph, value, **kw):
    return apply_list(dispgraph, value, kw)


if __name__ == '__main__': # pragma: no cover
    import doctest
    doctest.testmod()
